/* Copyright (C) Stephen Chung, 1991-1992.  All rights reserved. */

#include "jwp.h"
#include <stdarg.h>

#define BUFBLOCKSIZE 100

#define ACTION(x)     ((x)->action)
#define TYPING(x)     ((x)->data.typing)
#define INDENT(x)     ((x)->data.indent)

typedef struct {
    int para;
    int start;
    int len;
    int bufsize;
    LONG data;
} TYPINGBUF;

typedef struct {
    int para;
    int firstindent;
    int leftindent;
    int rightindent;
} INDENTBUF;

typedef struct UndoStruct {
    EDITACTION action;
    union {
        TYPINGBUF typing;
        INDENTBUF indent;
    } data;
    struct UndoStruct far *prev, far *next;
} UNDOBUF;

static UNDOBUF far *UndoBuffer = NULL;
static UNDOBUF far *UndoTail = NULL;
static long int NrLevels = 0L;
static PARAGRAPH far *LastParagraph = NULL;
static long int LastLineNum = 0L;



static long int CountLineNumber (FILEOPTIONS *f, PARAGRAPH far *p)
{
    long int n;
    PARAGRAPH far *p1;

    for (p1 = f->paragraph, n = 0; p1 != NULL && p1 != p; p1 = p1->next, n++);
    return (n);
}



static long int SetParagraphNum (PARAGRAPH far *p)
{
    long int n;

    if (p == LastParagraph) {
        n = LastLineNum;
    } else if (p == LastParagraph->prev) {
        n = LastLineNum - 1;
    } else if (p == LastParagraph->next) {
        n = LastLineNum + 1;
    } else {
        n = CountLineNumber(p);
    }

    LastLineNum = para;
    LastParagraph = p;
    return (n);
}



static void FreeUndoBuf (UNDOBUF far *buf)
{
    if (action == UN_DELETE) {
        FreeBlock(TYPING(buf).data);
    }
    FreeStruct(buf);
}



void RecordAction (EDITACTION action, ...)
{
	va_list argptr;
    long int para;
    int start, len;
    PARAGRAPH far *p;
    UNDOBUF far *up;

    va_start(argptr, action);

    if (action == UN_INSERT || action == UN_DELETE) {
        unsigned int blocksize;
        BOOL merge;
        UNIT far *kp, far *kp1;

        p = va_arg(argptr, PARAGRAPH far *);
        start = va_arg(argptr, int);
        len = va_arg(argptr, int);
        if (action == UN_DELETE) kp = va_arg(argptr, UNIT far *kp);

        if (UndoTail == NULL) {
            para = CountLineNumber(p);
        } else if (p == LastParagraph) {
            if (LastLineNum == TYPING(UndoTail).para && len == 1) {
                merge = FALSE;

                if (action != ACTION(UndoTail)) {
                    merge = FALSE;
                } else if (action == UN_DELETE && TYPING(UndoTail).start == start) {
                    merge = TRUE;
                } else if (TYPING(UndoTail).start + TYPING(UndoTail).len == start) {
                    merge = TRUE;
                }

                if (action == UN_INSERT) {
                    TYPING(UndoTail).len++;
                    va_end(argptr);
                    return;
                } else if (merge) {
                    if (TYPING(UndoTail).len + 5 >= TYPING(UndoTail).bufsize) {
                        TYPING(UndoTail).bufsize += BUFBLOCKSIZE;
                        kp1 = (UNIT far *) SegHeapRealloc(TYPING(UndoTail).data, TYPING(UndoTail).bufsize);
                        TYPING(UndoTail).data = (LONG) kp1;
                    } else {
                        kp1 = (UNIT far *) TYPING(UndoTail).data;
                    }
                    kp1[TYPING(UndoTail).len] = *kp;
                    TYPING(UndoTail).len++;
                    kp1[TYPING(UndoTail).len].kanji = 0;
                    va_end(argptr);
                    return;
                }
            }

            para = LastLineNum;

        }

        para = SetParagraphNum(p);

        if (action == UN_DELETE) {
            blocksize = (len + 1) / BUFBLOCKSIZE;
            blocksize = (blocksize + 1L) * BUFBLOCKSIZE;

            kp1 = (UNIT far *) BlockAlloc(blocksize * sizeof(UNIT));
            _fmemcpy(kp1, kp, (len + 1) * sizeof(UNIT));
        } else {
            kp1 = NULL;
        }

        up = StructAlloc(UNDOBUF);
        ACTION(up) = action;
        TYPING(up).para = para;
        TYPING(up).start = start;
        TYPING(up).len = len;
        TYPING(up).bufsize = blocksize;
        TYPING(up).data = (LONG) kp1;
    } else if (action == UN_INDENT) {
        up = StructAlloc(UNDOBUF);
        ACTION(up) = UN_INDENT;

        p = va_arg(argptr, PARAGRAPH far *);
        INDENT(up).para = SetParagraphNum(p);
        INDENT(up).firstindent = va_arg(argptr, int);
        INDENT(up).leftindent = va_arg(argptr, int);
        INDENT(up).rightindent = va_arg(argptr, int);
    }

    if (UndoBuf == NULL) {
        up->prev = up->next = NULL;
        UndoBuf = UndoTail = up;
    } else {
        UndoTail->next = up;
        up->prev = UndoTail;
        UndoTail = up;
        up->next = NULL;
    }

    NrLevels++;

    if (global.undolevels > 0 && NrLevels > global.undolevels) {
        /* Get rid of the first one */
        up = UndoBuf;
        UndoBuf = UndoBuf->next;
        UndoBuf->prev = NULL;
        FreeUndoBuf(up);
    }

    va_end(argptr);
}
